'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function overrideConfigFromArgv(config, argv) {
    if (argv.prod) {
        config.devMode = false;
    }
    else if (argv.dev) {
        config.devMode = true;
    }
    if (argv.watch) {
        config.watch = true;
    }
    if (argv.debug) {
        config.logLevel = 'debug';
    }
    else if (argv.logLevel) {
        config.logLevel = argv.logLevel;
    }
    if (!argv.prerender) {
        config.prerender = false;
    }
    else if (!config.prerender) {
        config.prerender = true;
    }
    if (config.devMode) {
        if (argv.serviceWorker && config.serviceWorker) {
            // dev mode, but forcing service worker
            // and they provided a sw config
            // so generate a service worker with their config
            config.serviceWorker = true;
        }
        else if (argv.serviceWorker && config.serviceWorker === undefined) {
            // dev mode, but forcing service worker
            // but they didn't provide a sw config
            // so still force it to generate w/ our defaults
            config.serviceWorker = true;
        }
        else {
            // service worker is off by default in dev mode
            config.serviceWorker = false;
        }
    }
    else if (config.serviceWorker === undefined) {
        // prod mode, and they didn't provide a sw config
        // so force it generate with our defaults
        config.serviceWorker = true;
    }
    if (argv.es5) {
        config.buildEs5 = true;
    }
    if (argv.docs) {
        config.generateDocs = true;
    }
}
function getConfigFilePath(process, sys, configArg) {
    if (configArg) {
        if (!sys.path.isAbsolute(configArg)) {
            // passed in a custom stencil config location
            // but it's relative, so prefix the cwd
            return sys.path.join(process.cwd(), configArg);
        }
        // config path already an absolute path, we're good here
        return configArg;
    }
    // nothing was passed in, use the current working directory
    return process.cwd();
}
function parseArgv(process) {
    const minimist = require('minimist');
    const cmdArgs = getCmdArgs(process);
    const argv = minimist(cmdArgs, ARG_OPTS);
    argv.serviceWorker = argv['service-worker'];
    argv.logLevel = argv['log-level'];
    return argv;
}
const ARG_OPTS = {
    boolean: [
        'debug',
        'dev',
        'docs',
        'es5',
        'help',
        'prod',
        'prerender',
        'service-worker',
        'skip-node-check',
        'version',
        'watch'
    ],
    string: [
        'config',
        'log-level'
    ],
    alias: {
        'c': 'config',
        'h': 'help',
        'v': 'version'
    }
};
function getCmdArgs(process) {
    let cmdArgs = process.argv.slice(2);
    try {
        var npmRunArgs = process.env.npm_config_argv;
        if (npmRunArgs) {
            cmdArgs = cmdArgs.concat(JSON.parse(npmRunArgs).original);
        }
    }
    catch (e) { }
    return cmdArgs;
}
function hasError(diagnostics) {
    if (!diagnostics) {
        return false;
    }
    return diagnostics.some(d => d.level === 'error' && d.type !== 'runtime');
}

function help(process, logger) {
    var p = logger.dim((process.platform === 'win32') ? '>' : '$');
    console.log(`
  ${logger.bold('Build:')} ${logger.dim('Build components for development or production.')}

    ${p} ${logger.green('stencil build [--dev] [--watch] [--prerender] [--debug]')}

      ${logger.green('--dev')} ${logger.dim('..................')} Execute a development build.
      ${logger.green('--watch')} ${logger.dim('................')} Execute a build in watch mode.
      ${logger.green('--prerender')} ${logger.dim('............')} Prerender URLs.
      ${logger.green('--debug')} ${logger.dim('................')} Set the log level to debug.
      ${logger.green('--config')} ${logger.dim('...............')} Stencil config file.
      ${logger.green('--docs')} ${logger.dim('.................')} Generate readme.md docs for each component

  ${logger.bold('Examples:')}

    ${p} ${logger.green('stencil build --dev --watch')}
    ${p} ${logger.green('stencil build --prerender')}
    ${p} ${logger.green('stencil init')}

`);
}

function initApp(process, sys, logger) {
    const configPath = sys.path.join(process.cwd(), 'stencil.config.js');
    try {
        sys.fs.writeFileSync(configPath, DEFAULT_CONFIG);
        logger.info(`Created config: ${configPath}`);
    }
    catch (e) {
        logger.error(e);
    }
}
var DEFAULT_CONFIG = `
exports.config = {
  namespace: 'App',
  bundles: [],
  collections: []
};

exports.devServer = {
  root: 'www',
  watchGlob: '**/**'
};
`;

function run(process, sys, logger, compiler) {
    const task = process.argv[2];
    const argv = parseArgv(process);
    process.on('unhandledRejection', (r) => logger.error(r));
    if (argv.help) {
        help(process, logger);
        return process.exit(0);
    }
    if (task === 'init') {
        initApp(process, sys, logger);
        return process.exit(0);
    }
    if (argv.version) {
        console.log(sys.compiler.version);
        return process.exit(0);
    }
    // load the config file
    let config;
    try {
        const configPath = getConfigFilePath(process, sys, argv.config);
        config = sys.loadConfigFile(configPath);
    }
    catch (e) {
        logger.error(e);
        return process.exit(1);
    }
    // override the config values with any cli arguments
    overrideConfigFromArgv(config, argv);
    if (!config.logger) {
        // if a logger was not provided then use the
        // default stencil command line logger
        config.logger = logger;
    }
    if (config.logLevel) {
        config.logger.level = config.logLevel;
    }
    if (!config.sys) {
        // if the config was not provided then use the default node sys
        config.sys = sys;
    }
    switch (task) {
        case 'build':
            compiler.build(config).then((results) => {
                if (!config.watch && hasError(results && results.diagnostics)) {
                    process.exit(1);
                }
            }).catch((err) => {
                config.logger.error(err);
                process.exit(1);
            });
            if (config.watch) {
                process.once('SIGINT', () => {
                    return process.exit(0);
                });
            }
            break;
        case 'docs':
            config.generateDocs = true;
            compiler.docs(config).catch((err) => {
                config.logger.error(err);
            });
            break;
        default:
            config.logger.error(`Invalid stencil command, please see the options below:`);
            help(process, logger);
            process.exit(1);
    }
}

exports.run = run;
