'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

class NodeLogger {
    constructor(opts = {}) {
        this._level = 'info';
        this.level = opts.level;
        const path = require('path');
        const sysUtil = require(path.join(__dirname, './sys-util.js'));
        this.chalk = sysUtil.chalk;
    }
    get level() {
        return this._level;
    }
    set level(l) {
        if (typeof l === 'string') {
            l = l.toLowerCase().trim();
            if (LOG_LEVELS.indexOf(l) === -1) {
                this.error(`Invalid log level '${this.chalk.bold(l)}' (choose from: ${LOG_LEVELS.map(l => this.chalk.bold(l)).join(', ')})`);
            }
            else {
                this._level = l;
            }
        }
    }
    info(...msg) {
        if (this.shouldLog('info')) {
            const lines = wordWrap(msg);
            this.infoPrefix(lines);
            console.log(lines.join('\n'));
        }
    }
    infoPrefix(lines) {
        if (lines.length) {
            const d = new Date();
            let prefix = '[' +
                ('0' + d.getMinutes()).slice(-2) + ':' +
                ('0' + d.getSeconds()).slice(-2) + '.' +
                Math.floor((d.getMilliseconds() / 1000) * 10) + ']';
            lines[0] = this.dim(prefix) + lines[0].substr(prefix.length);
        }
    }
    warn(...msg) {
        if (this.shouldLog('warn')) {
            const lines = wordWrap(msg);
            this.warnPrefix(lines);
            console.warn(lines.join('\n'));
        }
    }
    warnPrefix(lines) {
        if (lines.length) {
            let prefix = '[ WARN  ]';
            lines[0] = this.bold(this.chalk.yellow(prefix)) + lines[0].substr(prefix.length);
        }
    }
    error(...msg) {
        if (this.shouldLog('error')) {
            const lines = wordWrap(msg);
            this.errorPrefix(lines);
            console.error(lines.join('\n'));
        }
    }
    errorPrefix(lines) {
        if (lines.length) {
            let prefix = '[ ERROR ]';
            lines[0] = this.bold(this.chalk.red(prefix)) + lines[0].substr(prefix.length);
        }
    }
    debug(...msg) {
        if (this.shouldLog('debug')) {
            msg.push(this.memoryUsage());
            const lines = wordWrap(msg);
            this.debugPrefix(lines);
            console.log(lines.join('\n'));
        }
    }
    debugPrefix(lines) {
        if (lines.length) {
            const d = new Date();
            let prefix = '[' +
                ('0' + d.getMinutes()).slice(-2) + ':' +
                ('0' + d.getSeconds()).slice(-2) + '.' +
                Math.floor((d.getMilliseconds() / 1000) * 10) + ']';
            lines[0] = this.chalk.cyan(prefix) + lines[0].substr(prefix.length);
        }
    }
    color(msg, color) {
        return this.chalk[color](msg);
    }
    red(msg) {
        return this.chalk.red(msg);
    }
    green(msg) {
        return this.chalk.green(msg);
    }
    yellow(msg) {
        return this.chalk.yellow(msg);
    }
    blue(msg) {
        return this.chalk.blue(msg);
    }
    magenta(msg) {
        return this.chalk.magenta(msg);
    }
    cyan(msg) {
        return this.chalk.cyan(msg);
    }
    gray(msg) {
        return this.chalk.gray(msg);
    }
    bold(msg) {
        return this.chalk.bold(msg);
    }
    dim(msg) {
        return this.chalk.dim(msg);
    }
    memoryUsage() {
        return this.dim(` MEM: ${(process.memoryUsage().rss / 1000000).toFixed(1)}MB`);
    }
    shouldLog(level) {
        return LOG_LEVELS.indexOf(level) >= LOG_LEVELS.indexOf(this.level);
    }
    createTimeSpan(startMsg, debug = false) {
        return new CmdTimeSpan(this, startMsg, debug);
    }
    printDiagnostics(diagnostics) {
        if (!diagnostics || !diagnostics.length)
            return;
        let outputLines = [''];
        diagnostics.forEach(d => {
            outputLines = outputLines.concat(this.printDiagnostic(d));
        });
        console.log(outputLines.join('\n'));
    }
    printDiagnostic(d) {
        const outputLines = wordWrap([d.messageText]);
        if (d.header && d.header !== 'build error' && d.header !== 'build warn') {
            outputLines.unshift(INDENT + d.header);
        }
        outputLines.push('');
        if (d.lines && d.lines.length) {
            const lines = prepareLines(d.lines, 'text');
            lines.forEach(l => {
                if (!isMeaningfulLine(l.text)) {
                    return;
                }
                let msg = `L${l.lineNumber}:  `;
                while (msg.length < INDENT.length) {
                    msg = ' ' + msg;
                }
                let text = l.text;
                if (l.errorCharStart > -1) {
                    text = this.highlightError(text, l.errorCharStart, l.errorLength);
                }
                msg = this.dim(msg);
                if (d.language === 'javascript') {
                    msg += this.jsSyntaxHighlight(text);
                }
                else if (d.language === 'scss' || d.language === 'css') {
                    msg += this.cssSyntaxHighlight(text);
                }
                else {
                    msg += text;
                }
                outputLines.push(msg);
            });
            outputLines.push('');
        }
        if (d.level === 'error') {
            this.errorPrefix(outputLines);
        }
        else if (d.level === 'warn') {
            this.warnPrefix(outputLines);
        }
        else if (d.level === 'debug') {
            this.debugPrefix(outputLines);
        }
        else {
            this.infoPrefix(outputLines);
        }
        return outputLines;
    }
    highlightError(errorLine, errorCharStart, errorLength) {
        let rightSideChars = errorLine.length - errorCharStart + errorLength - 1;
        while (errorLine.length + INDENT.length > MAX_LEN) {
            if (errorCharStart > (errorLine.length - errorCharStart + errorLength) && errorCharStart > 5) {
                // larger on left side
                errorLine = errorLine.substr(1);
                errorCharStart--;
            }
            else if (rightSideChars > 1) {
                // larger on right side
                errorLine = errorLine.substr(0, errorLine.length - 1);
                rightSideChars--;
            }
            else {
                break;
            }
        }
        const lineChars = [];
        const lineLength = Math.max(errorLine.length, errorCharStart + errorLength);
        for (var i = 0; i < lineLength; i++) {
            var chr = errorLine.charAt(i);
            if (i >= errorCharStart && i < errorCharStart + errorLength) {
                chr = this.chalk.bgRed(chr === '' ? ' ' : chr);
            }
            lineChars.push(chr);
        }
        return lineChars.join('');
    }
    jsSyntaxHighlight(text) {
        if (text.trim().startsWith('//')) {
            return this.dim(text);
        }
        const words = text.split(' ').map(word => {
            if (JS_KEYWORDS.indexOf(word) > -1) {
                return this.chalk.cyan(word);
            }
            return word;
        });
        return words.join(' ');
    }
    cssSyntaxHighlight(text) {
        let cssProp = true;
        const safeChars = 'abcdefghijklmnopqrstuvwxyz-_';
        const notProp = '.#,:}@$[]/*';
        const chars = [];
        for (var i = 0; i < text.length; i++) {
            var c = text.charAt(i);
            if (c === ';' || c === '{') {
                cssProp = true;
            }
            else if (notProp.indexOf(c) > -1) {
                cssProp = false;
            }
            if (cssProp && safeChars.indexOf(c.toLowerCase()) > -1) {
                chars.push(this.chalk.cyan(c));
                continue;
            }
            chars.push(c);
        }
        return chars.join('');
    }
}
class CmdTimeSpan {
    constructor(logger, startMsg, debug) {
        this.debug = debug;
        this.logger = logger;
        this.start = Date.now();
        let msg = `${startMsg} ${logger.dim('...')}`;
        if (this.debug) {
            this.logger.debug(msg);
        }
        else {
            this.logger.info(msg);
        }
    }
    finish(msg, color, bold, newLineSuffix) {
        if (color) {
            msg = this.logger.color(msg, color);
        }
        if (bold) {
            msg = this.logger.bold(msg);
        }
        msg += ' ' + this.logger.dim(this.timeSuffix());
        if (this.debug) {
            this.logger.debug(msg);
        }
        else {
            this.logger.info(msg);
        }
        if (newLineSuffix) {
            console.log('');
        }
    }
    timeSuffix() {
        const duration = Date.now() - this.start;
        let time;
        if (duration > 1000) {
            time = 'in ' + (duration / 1000).toFixed(2) + ' s';
        }
        else {
            let ms = parseFloat((duration).toFixed(3));
            if (ms > 0) {
                time = 'in ' + duration + ' ms';
            }
            else {
                time = 'in less than 1 ms';
            }
        }
        return time;
    }
}
const LOG_LEVELS = ['debug', 'info', 'warn', 'error'];
function wordWrap(msg) {
    const lines = [];
    const words = [];
    msg.forEach(m => {
        if (m === null) {
            words.push('null');
        }
        else if (typeof m === 'undefined') {
            words.push('undefined');
        }
        else if (typeof m === 'string') {
            m.replace(/\s/gm, ' ').split(' ').forEach(strWord => {
                if (strWord.trim().length) {
                    words.push(strWord.trim());
                }
            });
        }
        else if (typeof m === 'number' || typeof m === 'boolean' || typeof m === 'function') {
            words.push(m.toString());
        }
        else if (Array.isArray(m)) {
            words.push(() => {
                return m.toString();
            });
        }
        else if (Object(m) === m) {
            words.push(() => {
                return m.toString();
            });
        }
        else {
            words.push(m.toString());
        }
    });
    let line = INDENT;
    words.forEach(word => {
        if (lines.length > 25) {
            return;
        }
        if (typeof word === 'function') {
            if (line.trim().length) {
                lines.push(line);
            }
            lines.push(word());
            line = INDENT;
        }
        else if (INDENT.length + word.length > MAX_LEN) {
            // word is too long to play nice, just give it its own line
            if (line.trim().length) {
                lines.push(line);
            }
            lines.push(INDENT + word);
            line = INDENT;
        }
        else if ((word.length + line.length) > MAX_LEN) {
            // this word would make the line too long
            // print the line now, then start a new one
            lines.push(line);
            line = INDENT + word + ' ';
        }
        else {
            line += word + ' ';
        }
    });
    if (line.trim().length) {
        lines.push(line);
    }
    return lines;
}
function prepareLines(orgLines, code) {
    const lines = JSON.parse(JSON.stringify(orgLines));
    for (let i = 0; i < 100; i++) {
        if (!eachLineHasLeadingWhitespace(lines, code)) {
            return lines;
        }
        for (let i = 0; i < lines.length; i++) {
            lines[i][code] = lines[i][code].substr(1);
            lines[i].errorCharStart--;
            if (!lines[i][code].length) {
                return lines;
            }
        }
    }
    return lines;
}
function eachLineHasLeadingWhitespace(lines, code) {
    if (!lines.length) {
        return false;
    }
    for (var i = 0; i < lines.length; i++) {
        if (!lines[i][code] || lines[i][code].length < 1) {
            return false;
        }
        var firstChar = lines[i][code].charAt(0);
        if (firstChar !== ' ' && firstChar !== '\t') {
            return false;
        }
    }
    return true;
}
function isMeaningfulLine(line) {
    if (line) {
        line = line.trim();
        if (line.length) {
            return (MEH_LINES.indexOf(line) < 0);
        }
    }
    return false;
}
const MEH_LINES = [';', ':', '{', '}', '(', ')', '/**', '/*', '*/', '*', '({', '})'];
const JS_KEYWORDS = [
    'abstract', 'any', 'as', 'break', 'boolean', 'case', 'catch', 'class',
    'console', 'const', 'continue', 'debugger', 'declare', 'default', 'delete',
    'do', 'else', 'enum', 'export', 'extends', 'false', 'finally', 'for', 'from',
    'function', 'get', 'if', 'import', 'in', 'implements', 'Infinity',
    'instanceof', 'let', 'module', 'namespace', 'NaN', 'new', 'number', 'null',
    'public', 'private', 'protected', 'require', 'return', 'static', 'set',
    'string', 'super', 'switch', 'this', 'throw', 'try', 'true', 'type',
    'typeof', 'undefined', 'var', 'void', 'with', 'while', 'yield',
];
const INDENT = '           ';
const MAX_LEN = 120;

function createContext(ctx, wwwDir, sandbox) {
    const vm = require('vm');
    // https://github.com/tmpvar/jsdom/issues/1724
    // manually adding a fetch polyfill until jsdom adds it
    patchFetch(ctx, wwwDir, sandbox);
    patchRaf(sandbox);
    return vm.createContext(sandbox);
}
function patchFetch(ctx, wwwDir, sandbox) {
    function fetch(input, init) {
        var path = require('path');
        var nf = require(path.join(__dirname, './node-fetch.js'));
        createServer(ctx, wwwDir);
        if (typeof input === 'string') {
            // fetch(url)
            return nf.nodeFetch(normalizeUrl(input), init);
        }
        else {
            // fetch(Request)
            input.url = normalizeUrl(input.url);
            return nf.nodeFetch(input, init);
        }
    }
    sandbox.fetch = fetch;
}
function normalizeUrl(url) {
    var Url = require('url');
    var parsedUrl = Url.parse(url);
    if (!parsedUrl.protocol || !parsedUrl.hostname) {
        parsedUrl.protocol = 'http:';
        parsedUrl.host = 'localhost:' + PORT;
        url = Url.format(parsedUrl);
    }
    return url;
}
function patchRaf(sandbox) {
    if (!sandbox.requestAnimationFrame) {
        sandbox.requestAnimationFrame = function (callback) {
            var id = sandbox.setTimeout(function () {
                callback(Date.now());
            }, 0);
            return id;
        };
        sandbox.cancelAnimationFrame = function (id) {
            clearTimeout(id);
        };
    }
}
function createServer(ctx, wwwDir) {
    if (ctx.localPrerenderServer)
        return;
    var fs = require('fs');
    var path = require('path');
    var http = require('http');
    var Url = require('url');
    ctx.localPrerenderServer = http.createServer((request, response) => {
        var parsedUrl = Url.parse(request.url);
        var filePath = path.join(wwwDir, parsedUrl.pathname);
        fs.readFile(filePath, 'utf-8', (err, data) => {
            if (err) {
                response.write('Error fetching: ' + parsedUrl.pathname + ' : ' + err);
            }
            else {
                response.write(data);
            }
            response.end();
        });
    });
    ctx.localPrerenderServer.listen(PORT);
}
var PORT = 53536;
function runInContext(code, contextifiedSandbox, options) {
    const vm = require('vm');
    vm.runInContext(code, contextifiedSandbox, options);
}

function createDom() {
    let dom = null;
    return {
        parse(opts) {
            if (dom) {
                dom.window.close();
            }
            const jsdom = require('jsdom');
            const jsdomOptions = {
                url: opts.url,
                referrer: opts.referrer,
                userAgent: opts.userAgent
            };
            if (opts.console) {
                jsdomOptions.virtualConsole = new jsdom.VirtualConsole();
                jsdomOptions.virtualConsole.sendTo(opts.console);
            }
            dom = new jsdom.JSDOM(opts.html, jsdomOptions);
            return dom.window;
        },
        serialize() {
            return dom.serialize();
        },
        destroy() {
            dom.window.close();
            dom = null;
        }
    };
}

/**
 * SSR Attribute Names
 */


/**
 * Default style mode id
 */

/**
 * Reusable empty obj/array
 * Don't add values to these!!
 */


/**
 * Key Name to Key Code Map
 */

/**
 * Namespaces
 */



/**
 * File names and value
 */

var __awaiter = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};



















function normalizePath(str) {
    // Convert Windows backslash paths to slash paths: foo\\bar ➔ foo/bar
    // https://github.com/sindresorhus/slash MIT
    // By Sindre Sorhus
    if (EXTENDED_PATH_REGEX.test(str) || NON_ASCII_REGEX.test(str)) {
        return str;
    }
    return str.replace(SLASH_REGEX, '/');
}
const EXTENDED_PATH_REGEX = /^\\\\\?\\/;
const NON_ASCII_REGEX = /[^\x00-\x80]+/;
const SLASH_REGEX = /\\/g;

class NodeSystem {
    constructor(fs, path) {
        this.coreClientFileCache = {};
        this.nodeFs = fs || require('fs');
        this.nodePath = path || require('path');
        this.sysUtil = require(this.nodePath.join(__dirname, './sys-util.js'));
        this.init();
    }
    init() {
        const packageRootDir = this.nodePath.join(__dirname, '../../../');
        this.packageDistDir = this.nodePath.join(packageRootDir, 'dist', 'client');
        try {
            this.packageJsonData = require(this.nodePath.join(packageRootDir, 'package.json'));
        }
        catch (e) {
            throw new Error(`unable to resolve "package.json" from: ${packageRootDir}`);
        }
        try {
            this.typescriptPackageJson = require(this.resolveModule(packageRootDir, 'typescript'));
        }
        catch (e) {
            throw new Error(`unable to resolve "typescript" from: ${packageRootDir}`);
        }
    }
    get compiler() {
        return {
            name: this.packageJsonData.name,
            version: this.packageJsonData.version,
            typescriptVersion: this.typescriptPackageJson.version
        };
    }
    copy(src, dest, opts) {
        return new Promise((resolve, reject) => {
            opts = opts || {};
            this.sysUtil.fsExtra.copy(src, dest, opts, (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    get createDom() {
        return createDom;
    }
    emptyDir(dir) {
        return new Promise((resolve, reject) => {
            this.sysUtil.fsExtra.emptyDir(dir, (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    ensureDir(dir) {
        return new Promise((resolve, reject) => {
            this.sysUtil.fsExtra.ensureDir(dir, (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    ensureDirSync(dir) {
        this.sysUtil.fsExtra.ensureDirSync(dir);
    }
    ensureFile(file) {
        return new Promise((resolve, reject) => {
            this.sysUtil.fsExtra.ensureFile(file, (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    get fs() {
        return this.nodeFs;
    }
    generateContentHash(content, length) {
        const crypto = require('crypto');
        return crypto.createHash('sha1')
            .update(content)
            .digest('base64')
            .replace(/\W/g, '')
            .substr(0, length)
            .toLowerCase();
    }
    getClientCoreFile(opts) {
        const filePath = this.nodePath.join(this.packageDistDir, opts.staticName);
        return new Promise((resolve, reject) => {
            if (this.coreClientFileCache[filePath]) {
                resolve(this.coreClientFileCache[filePath]);
            }
            else {
                this.nodeFs.readFile(filePath, 'utf-8', (err, data) => {
                    if (err) {
                        reject(err);
                    }
                    else {
                        this.coreClientFileCache[filePath] = data;
                        resolve(data);
                    }
                });
            }
        });
    }
    glob(pattern, opts) {
        return new Promise((resolve, reject) => {
            this.sysUtil.glob(pattern, opts, (err, files) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve(files);
                }
            });
        });
    }
    isGlob(str) {
        return this.sysUtil.isGlob(str);
    }
    loadConfigFile(configPath) {
        let config;
        if (!this.nodePath.isAbsolute(configPath)) {
            throw new Error(`Stencil configuration file "${configPath}" must be an absolute path.`);
        }
        try {
            const fileStat = this.nodeFs.statSync(configPath);
            if (fileStat.isDirectory()) {
                // this is only a directory, so let's just assume we're looking for in stencil.config.js
                // otherwise they could pass in an absolute path if it was somewhere else
                configPath = this.nodePath.join(configPath, 'stencil.config.js');
            }
            // the passed in config was a string, so it's probably a path to the config we need to load
            const configFileData = require(configPath);
            if (!configFileData.config) {
                throw new Error(`Invalid Stencil configuration file "${configPath}". Missing "config" property.`);
            }
            config = configFileData.config;
            config.configPath = configPath;
            if (!config.rootDir && configPath) {
                config.rootDir = this.nodePath.dirname(configPath);
            }
        }
        catch (e) {
            throw new Error(`Error reading Stencil configuration file "${configPath}". ` + e);
        }
        if (!config.sys) {
            config.sys = this;
        }
        return config;
    }
    minifyCss(input) {
        const CleanCSS = require(this.nodePath.join(__dirname, './clean-css.js')).cleanCss;
        const result = new CleanCSS().minify(input);
        const diagnostics = [];
        if (result.errors) {
            result.errors.forEach((msg) => {
                diagnostics.push({
                    header: 'Minify CSS',
                    messageText: msg,
                    level: 'error',
                    type: 'build'
                });
            });
        }
        if (result.warnings) {
            result.warnings.forEach((msg) => {
                diagnostics.push({
                    header: 'Minify CSS',
                    messageText: msg,
                    level: 'warn',
                    type: 'build'
                });
            });
        }
        return {
            output: result.styles,
            sourceMap: result.sourceMap,
            diagnostics: diagnostics
        };
    }
    minifyJs(input, opts) {
        const UglifyJS = require('uglify-es');
        const result = UglifyJS.minify(input, opts);
        const diagnostics = [];
        if (result.error) {
            diagnostics.push({
                header: 'Minify JS',
                messageText: result.error.message,
                level: 'error',
                type: 'build'
            });
        }
        return {
            output: result.code,
            sourceMap: result.sourceMap,
            diagnostics: diagnostics
        };
    }
    minimatch(filePath, pattern, opts) {
        return this.sysUtil.minimatch(filePath, pattern, opts);
    }
    get path() {
        return this.nodePath;
    }
    remove(dir) {
        return new Promise((resolve, reject) => {
            this.sysUtil.fsExtra.remove(dir, (err) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    resolveModule(fromDir, moduleId) {
        const Module = require('module');
        fromDir = this.nodePath.resolve(fromDir);
        const fromFile = this.nodePath.join(fromDir, 'noop.js');
        let dir = Module._resolveFilename(moduleId, {
            id: fromFile,
            filename: fromFile,
            paths: Module._nodeModulePaths(fromDir)
        });
        const root = this.nodePath.parse(fromDir).root;
        let packageJsonFilePath;
        while (dir !== root) {
            dir = this.nodePath.dirname(dir);
            packageJsonFilePath = this.nodePath.join(dir, 'package.json');
            try {
                this.nodeFs.accessSync(packageJsonFilePath);
            }
            catch (e) {
                continue;
            }
            return normalizePath(packageJsonFilePath);
        }
        throw new Error(`error loading "${moduleId}" from "${fromDir}"`);
    }
    get rollup() {
        const rollup = require('rollup');
        rollup.plugins = {
            commonjs: require('rollup-plugin-commonjs'),
            nodeResolve: require('rollup-plugin-node-resolve')
        };
        return rollup;
    }
    get sass() {
        return require('node-sass');
    }
    get semver() {
        return this.sysUtil.semver;
    }
    get typescript() {
        return require('typescript');
    }
    get url() {
        return require('url');
    }
    get vm() {
        return {
            createContext,
            runInContext
        };
    }
    watch(paths, opts) {
        const chokidar = require('chokidar');
        return chokidar.watch(paths, opts);
    }
    get workbox() {
        return require('workbox-build');
    }
}

exports.NodeLogger = NodeLogger;
exports.NodeSystem = NodeSystem;
